# module for loading parameters.
import os

def loadParametersD(paramFN):
    '''Load a parameters file into a dictionary. Assumes the file consists
of python assignment statements.'''
    
    paramD={}
    f=open(paramFN,'r')

    while True:
        s=f.readline()
        if s == '':
            break
        if s[0]=='#' or s[0] == '\n':
            continue
        # if it's not the EOF, a comment, or a blank line, so it
        # must be an assignment statement.
        key,value = s.rstrip().split('=')
        key = key.strip()
        value = value.strip()
        paramD[key] = eval(value)

    f.close()

    return paramD

def loadFileNameMapD(fileNameMapFN,genbankFileList=None):
    '''Create a dictionary with mappings between genbank file names and
the human readable names we use in the tree. If fileNameMapFN contains
a string, we load that file and construct the mappings based on this.
Expects file with one species per line: genbank name + white space +
human name. If fileNameMapFN is None, we create the mappings between
the full file names (with path and extension) and the stem of the file
name, which in this case should correspond to what is in the input
tree.
    '''
    fileNameMapD = {}
    if fileNameMapFN == None:
        for fullPathFN in genbankFileList:
            fn = os.path.split(fullPathFN)[-1]
            stem = os.path.splitext(fn)[0]
            fileNameMapD[fn] = stem
    else:
        f = open(fileNameMapFN,'r')
        while True:
            s = f.readline()
            if s == '':
                break
            elif s[0].isspace():
                continue
            genbankStem,human = s.rstrip().split()
            fileNameMapD[genbankStem] = human
    return fileNameMapD



"""Provides the entry point to xenoGI's functionality."""
__version__ = "1.1.2"
import sys, glob, os
from . import parameters,genbank,blast,trees,genomes,Score,scores,Family,families,islands,analysis,islandBed

def main():
    
    #### check command line
    try:
        assert(len(sys.argv) == 3)
        paramFN=sys.argv[1]
        task = sys.argv[2]
        assert(task in ['parseGenbank', 'runBlast', 'calcScores', 'makeFamilies', 'makeIslands', 'printAnalysis', 'createIslandBed', 'interactiveAnalysis', 'runAll'])
    
    except:
        print(
            """
   Exactly two arguments required.
      1. A path to a parameter file.
            2. The task to be run which must be one of: parseGenbank, runBlast, calcScores, makeFamilies, makeIslands, printAnalysis, createIslandBed, interactiveAnalysis, or runAll.

   For example: 
      xenoGI params.py parseGenbank
"""
            ,file=sys.stderr)
        sys.exit(1)

        
    #### load parameters and some other data we'll use below
    paramD = parameters.loadParametersD(paramFN)
        
    #### parseGenbank
    if task == 'parseGenbank':
        parseGenbankWrapper(paramD)
        
    #### runBlast
    elif task == 'runBlast':
        blast.runBlast(paramD['fastaFilePath'],paramD['blastFilePath'],paramD['blastExecutDirPath'],paramD['blastCLine'],paramD['numThreads'],paramD['treeFN'])

    #### calcScores
    elif task == 'calcScores':
        calcScoresWrapper(paramD)

    #### makeFamilies
    elif task == 'makeFamilies':
        makeFamiliesWrapper(paramD)
        
    #### makeIslands
    elif task == 'makeIslands':
        makeIslandsWrapper(paramD)

    #### printAnalysis
    elif task == 'printAnalysis':
        printAnalysisWrapper(paramD)

    #### createIslandBed
    elif task == 'createIslandBed':
        createIslandBedWrapper(paramD)

    #### runAll
    elif task == 'runAll':
        parseGenbankWrapper(paramD)
        blast.runBlast(paramD['fastaFilePath'],paramD['blastFilePath'],paramD['blastExecutDirPath'],paramD['blastCLine'],paramD['numThreads'],paramD['treeFN'])
        calcScoresWrapper(paramD)
        makeFamiliesWrapper(paramD)
        makeIslandsWrapper(paramD)
        printAnalysisWrapper(paramD)
        createIslandBedWrapper(paramD)

    #### interactiveAnalysis
    elif task == 'interactiveAnalysis':
        interactiveAnalysisWrapper(paramD)

        
######## Task related functions

def parseGenbankWrapper(paramD):
    """Wrapper running stuff to parse genbank files."""
    genbankFileList=glob.glob(paramD['genbankFilePath'])

    # if directory for fastas doesn't exist yet, make it
    fastaDir = paramD['fastaFilePath'].split('*')[0]
    if glob.glob(fastaDir)==[]:
        os.mkdir(fastaDir)

    fileNameMapD = parameters.loadFileNameMapD(paramD['fileNameMapFN'],genbankFileList)

    # parse
    genbank.parseGenbank(paramD['geneOrderFN'],paramD['redundProtsFN'],paramD['geneInfoFN'],fastaDir,genbankFileList,fileNameMapD)

def loadMiscDataStructures(paramD):
    """Creates a few data structures that are used in multiple tasks."""

    tree,strainStr2NumD,strainNum2StrD = trees.readTree(paramD['treeFN'])
    
    # an object for gene name conversions
    geneNames = genomes.geneNames(paramD['geneOrderFN'],strainStr2NumD,strainNum2StrD)

    subtreeL=trees.createSubtreeL(tree)
    subtreeL.sort()
    geneOrderT=genomes.createGeneOrderTs(paramD['geneOrderFN'],geneNames,subtreeL,strainStr2NumD)

    return tree,strainStr2NumD,strainNum2StrD,geneNames,subtreeL,geneOrderT



def calcScoresWrapper(paramD):
    """Wrapper running stuff to calculate scores."""

    tree,strainStr2NumD,strainNum2StrD,geneNames,subtreeL,geneOrderT = loadMiscDataStructures(paramD)

    # object for storing scores
    scoresO=Score.Score()
    scoresO.initializeDataAttributes(paramD['blastFilePath'],geneNames)

    ## similarity scores
    scoresO = scores.calcRawScores(paramD['fastaFilePath'],paramD['numThreads'],geneNames,paramD['gapOpen'],paramD['gapExtend'],paramD['matrix'],scoresO)

    ## normalized scores
    scoresO,aabrhL,aabrhRawScoreSummmaryD=scores.calcNormScores(tree,strainNum2StrD,paramD['blastFilePath'],paramD['evalueThresh'],scoresO,geneNames,paramD['aabrhFN'])

    ## synteny scores
    scoresO = scores.calcSynScores(scoresO,aabrhRawScoreSummmaryD,geneNames,geneOrderT,paramD['synWSize'],paramD['numSynToTake'],paramD['numThreads'])

    ## core synteny scores
    scoresO = scores.calcCoreSynScores(scoresO,aabrhL,geneNames,geneOrderT,paramD['coreSynWsize'])

    # write scores to file
    scores.writeScores(scoresO,geneNames,paramD['scoresFN'])
    
def makeFamiliesWrapper(paramD):
    """Wrapper to create gene families."""

    tree,strainStr2NumD,strainNum2StrD,geneNames,subtreeL,geneOrderT = loadMiscDataStructures(paramD)
    
    ## read scores
    scoresO = scores.readScores(paramD['scoresFN'],geneNames)

    ## make gene families
    familyFormationSummaryF = open(paramD['familyFormationSummaryFN'],'w')
    familyL = families.families(tree,subtreeL,geneNames,scoresO,paramD['minNormThresh'],paramD['minCoreSynThresh'],paramD['minSynThresh'],paramD['synAdjustThresh'],paramD['synAdjustExtent'],paramD['familyFN'],strainNum2StrD,familyFormationSummaryF)
    familyFormationSummaryF.close()

def makeIslandsWrapper(paramD):
    """Wrapper to create islands"""

    tree,strainStr2NumD,strainNum2StrD,geneNames,subtreeL,geneOrderT = loadMiscDataStructures(paramD)
    
    ## read scores
    scoresO = scores.readScores(paramD['scoresFN'],geneNames)

    ## read gene families
    familyL = families.readFamilies(paramD['familyFN'],tree,geneNames,strainStr2NumD)

    ## group gene families into islands
    islandFormationSummaryF = open(paramD['islandFormationSummaryFN'],'w')
    islands.makeIslands(geneOrderT,geneNames,subtreeL,tree,paramD['proxThreshL'],familyL,paramD['numThreads'],strainStr2NumD,strainNum2StrD,paramD['rootFocalClade'],paramD['islandOutFN'],islandFormationSummaryF)
    islandFormationSummaryF.close()

def printAnalysisWrapper(paramD):
    """Wrapper to run analysis."""

    ## setup output directory and file names
    # if directory for analysis doesn't exist yet, make it
    analDir = paramD['analysisFilePath'].split("*")[0]
    if glob.glob(analDir)==[]:
        os.mkdir(analDir)

    islandSummaryStem = paramD['islandsSummaryStem']
    analExtension = paramD['analysisFilePath'].split("*")[1]
    islandsSummaryFN = os.path.join(analDir,islandSummaryStem+analExtension)

    genesFNstem = os.path.join(analDir,paramD['genesFNstem'])
    
    ## load stuff
    tree,strainStr2NumD,strainNum2StrD,geneNames,subtreeL,geneOrderT = loadMiscDataStructures(paramD)
    islandByNodeL=islands.readIslands(paramD['islandOutFN'],tree,strainStr2NumD)

    nodesL=trees.nodeList(tree)
    
    geneInfoD = genomes.readGeneInfoD(paramD['geneInfoFN'])

    familyL = families.readFamilies(paramD['familyFN'],tree,geneNames,strainStr2NumD)

    gene2FamD=analysis.createGene2FamD(familyL)
    fam2IslandD=analysis.createFam2IslandD(islandByNodeL)

    # Print out all islands
    islandsOutF = open(islandsSummaryFN,'w')
    analysis.vPrintAllIslands(islandByNodeL,tree,paramD['rootFocalClade'],subtreeL,familyL,strainStr2NumD,strainNum2StrD,geneNames,geneInfoD,islandsOutF)
    islandsOutF.close()

    ## analysis

    # scores
    scoresO = scores.readScores(paramD['scoresFN'],geneNames)
    scoresO.createNodeConnectL(geneNames) # make nodeConnectL attribute
    
    # calc family error scores
    families.calcErrorScores(familyL,scoresO,paramD['minNormThresh'],paramD['minCoreSynThresh'],paramD['minSynThresh'],paramD['famErrorScoreIncrementD'])

    # Print species files with all the genes, grouped by contig
    analysis.printSpeciesContigs(geneOrderT,genesFNstem,analExtension,geneNames,gene2FamD,fam2IslandD,geneInfoD,familyL,strainNum2StrD)

def createIslandBedWrapper(paramD):
    """Wrapper to make output bed files."""

    tree,strainStr2NumD,strainNum2StrD,geneNames,subtreeL,geneOrderT = loadMiscDataStructures(paramD)
    
    leafNodesL = trees.leafList(tree)
    geneNames = genomes.geneNames(paramD['geneOrderFN'],strainStr2NumD,strainNum2StrD)
    familyL = families.readFamilies(paramD['familyFN'],tree,geneNames,strainStr2NumD)
    islandByNodeL=islands.readIslands(paramD['islandOutFN'],tree,strainStr2NumD)
    geneInfoD = genomes.readGeneInfoD(paramD['geneInfoFN'])    


    # get islands organized by strain
    islandByStrainD = islandBed.createIslandByStrainD(leafNodesL,strainNum2StrD,islandByNodeL,familyL,geneNames,geneInfoD)

    islandBed.createAllBeds(islandByStrainD,geneInfoD,tree,strainNum2StrD,paramD['bedFilePath'],paramD['scoreNodeMapD'],paramD['potentialRgbL'],paramD['bedNumTries'])

def interactiveAnalysisWrapper(paramD):
    """Enter interactive mode."""

    ## Set up the modules a bit differently for interactive mode
    import code,sys
    from xenoGI.analysis import createGene2FamD,createFam2IslandD,printScoreMatrix,matchFamilyIsland,printIslandNeighb,vPrintIslands,coreNonCoreCtAtNode,printOutsideFamilyScores

    ## Wrapper analysis functions. For convenience these assume a
    ## bunch of global variables.
    
    def printFam(familyNum,fileF=sys.stdout):
        '''This is a wrapper to provide an easy way to print relevant info on
    a family. For ease of use, we take only one argument, assuming all the
    other required stuff is available at the top level. familyNum is the
    numerical identifier of a family.
        '''

        print("Family error score (count of possibly misassigned genes):",familyL[familyNum].possibleErrorCt,file=fileF)

        print(file=fileF)
        print("Matrix of raw similarity scores [0,1] between genes in the family",file=fileF)
        printScoreMatrix(familyNum,subtreeL,familyL,geneNames,scoresO,'rawSc',fileF)
        print(file=fileF)
        print(file=fileF)

        print(file=fileF)
        print("Matrix of normalized similarity scores between genes in the family",file=fileF)
        printScoreMatrix(familyNum,subtreeL,familyL,geneNames,scoresO,'normSc',fileF)
        print(file=fileF)
        print(file=fileF)

        print("Matrix of core synteny scores [0,1] between genes in the family",file=fileF)
        printScoreMatrix(familyNum,subtreeL,familyL,geneNames,scoresO,'coreSynSc',fileF)
        print(file=fileF)
        print(file=fileF)

        print("Matrix of synteny scores between genes in the family",file=fileF)
        printScoreMatrix(familyNum,subtreeL,familyL,geneNames,scoresO,'synSc',fileF)
        print(file=fileF)
        print(file=fileF)

        printOutsideFamilyScores(familyNum,subtreeL,familyL,geneNames,scoresO,fileF)
        print(file=fileF)
        print(file=fileF)

    def findIsland(searchStr,fileF=sys.stdout):
        '''Print the gene, family and island associated with searchStr. This
    is a wrapper that assumes various required objects are present at the
    top level.'''
        L=matchFamilyIsland(geneInfoD,geneNames,gene2FamD,fam2IslandD,searchStr)
        for geneName,fam,isl in L:
            print("<gene:"+str(geneName)+">","<family:"+str(fam)+">","<island:"+str(isl)+">",file=fileF)


    def printIsland(islandNum,synWSize,fileF=sys.stdout):
        '''Print the island and its genomic context in each species. We
        include synWSize/2 genes in either direction beyond the island.
        '''
        printIslandNeighb(islandNum,synWSize,subtreeL,islandByNodeL,familyL,geneOrderT,gene2FamD,fam2IslandD,geneInfoD,geneNames,strainNum2StrD,fileF)


    def printIslandsAtNode(nodeStr,fileF=sys.stdout):
        '''This is a wrapper to provide an easy way to print all the islands
    at a particular node in the tree. For ease of use, we take only a node
    number as argument, assuming all the other required stuff is available
    at the top level.
        '''
        node = strainStr2NumD[nodeStr]
        vPrintIslands(islandByNodeL[node],subtreeL,familyL,strainNum2StrD,geneNames,geneInfoD,fileF)


    def printCoreNonCoreByNode(fileF=sys.stdout):
        '''For each node in the focal clade, print the number of core and
    non-core families. That is, for that node, we get all families present
    in descendent species. We then look at their mrcas. Those with an mrca
    below the node in question are non-core, others are core.'''

        focTree = trees.subtree(tree,strainStr2NumD[paramD['rootFocalClade']])
        focInodesL=trees.iNodeList(focTree)
        familyByNodeL=createFamilyByNodeL(geneOrderT,gene2FamD)

        rowL=[]
        rowL.append(['Node','Core','Non-Core','Total','% Non-Core'])
        rowL.append(['----','----','--------','-----','----------'])
        for node in focInodesL:
            nonCore,core=coreNonCoreCtAtNode(tree,node,familyByNodeL,familyL)
            rowL.append([strainNum2StrD[node],str(core),str(nonCore),str(core+nonCore),str(format(nonCore/(core+nonCore),".3f"))])
        printTable(rowL,fileF=fileF)
        print(file=fileF)
        return


    ## Load data
    
    tree,strainStr2NumD,strainNum2StrD,geneNames,subtreeL,geneOrderT = loadMiscDataStructures(paramD)
    nodesL=trees.nodeList(tree)
    geneNames = genomes.geneNames(paramD['geneOrderFN'],strainStr2NumD,strainNum2StrD)
    geneInfoD = genomes.readGeneInfoD(paramD['geneInfoFN'])
    familyL = families.readFamilies(paramD['familyFN'],tree,geneNames,strainStr2NumD)
    islandByNodeL=islands.readIslands(paramD['islandOutFN'],tree,strainStr2NumD)
    gene2FamD=createGene2FamD(familyL)
    fam2IslandD=createFam2IslandD(islandByNodeL)
    geneOrderT=genomes.createGeneOrderTs(paramD['geneOrderFN'],geneNames,subtreeL,strainStr2NumD)
    scoresO = scores.readScores(paramD['scoresFN'],geneNames)
    scoresO.createNodeConnectL(geneNames) # make nodeConnectL attribute
    # calc family error scores
    families.calcErrorScores(familyL,scoresO,paramD['minNormThresh'],paramD['minCoreSynThresh'],paramD['minSynThresh'],paramD['famErrorScoreIncrementD'])
    
    code.interact(local=locals())




import sys
from multiprocessing import Pool
from . import trees
from . import genomes
from . import analysis
from .Island import *
from .Family import *

## Main function

def makeIslands(geneOrderT,geneNames,subtreeL,tree,proxThreshL,familyL,numThreads,strainStr2NumD,strainNum2StrD,rootFocalClade,islandOutFN, outputSummaryF):
    '''Parallelized wrapper to merge islands at different nodes.'''

    maxGeneProximityForIsland = max([thresh for thresh,rsc in proxThreshL])
    geneProximityD = genomes.createGeneProximityD(geneOrderT,maxGeneProximityForIsland)
    islandByNodeL=createIslandL(familyL,tree)

    focalSubtree = trees.subtree(tree,strainStr2NumD[rootFocalClade])
    focalNodesL=trees.nodeList(focalSubtree)
    

    ## create argumentL to be passed to p.map and mergeIslandsAtNode
    argumentL = []
    for mrcaNode in range(len(islandByNodeL)):
        if mrcaNode in focalNodesL:
            argumentL.append((islandByNodeL[mrcaNode],geneProximityD,proxThreshL,subtreeL[mrcaNode],familyL))
            
    # run it
    p=Pool(numThreads)
    islandByNodeLMerged = p.map(mergeIslandsAtNode, argumentL) 

    # islandByNodeLMerged has all islands from focal clade. Now add in
    # the single family islands from other nodes
    for mrcaNode in range(len(islandByNodeL)):
        if mrcaNode not in focalNodesL: 
            islandByNodeLMerged.append(islandByNodeL[mrcaNode])
            
    print("Merging complete.",file=sys.stderr)

    # print summary of merging
    printSummary(islandByNodeLMerged,focalNodesL,strainNum2StrD,islandByNodeL,outputSummaryF)
    
    # write islands
    writeIslands(islandByNodeLMerged,strainNum2StrD,islandOutFN)

    print("Islands written.",file=sys.stderr)

## Support functions

def createIslandL(familyL,tree):
    '''Greate islands, one family per island. Initially store islands
separately by mrca in a list of lists. (The index of the outer list
corresponds to the mrca)
    '''
    islandL=[[] for i in range(trees.nodeCount(tree))]

    for fam in familyL:
        isl = Island(fam.id, fam.mrca, [fam.id])
        islandL[isl.mrca].append(isl)

    # sort each list by island number
    for i in range(len(islandL)):
        islandL[i].sort(key=lambda x: x.id)
    
    return islandL

            
## Distance calculations

def proximity(fam1,fam2,geneProximityD,proximityThreshold,node):
    '''Return True if any of the genes at node from fam1 are within
proximityThreshold of genes at that node for fam2. proximityThreshold
is measured in genes, e.g. 1 means the adjacent gene.
    '''
    for gn1 in fam1.famGeneT[node]:
        for gn2 in fam2.famGeneT[node]:
            if gn1<gn2: key = (gn1,gn2)
            else: key = (gn2,gn1)
            if key in geneProximityD and geneProximityD[key] <= proximityThreshold:
                return True
    return False
  
def rcost(fam1,fam2,geneProximityD,proximityThreshold,subtree,rootProximate):
    '''Parsimony based calculation of cost of evolutionary rearrangement
given fam1 and fam2 begin at root of subtree. Assume either proximate
(nearby) or not at root (specificed by rootProximate). Charge 1 for
each rearrangment.
    '''
    if subtree[1]==():
        prox=proximity(fam1,fam2,geneProximityD,proximityThreshold,subtree[0])
        if (prox and rootProximate) or (not prox and not rootProximate):
            # state given by rootProximate matches adjacency info in our data
            return 0
        else:
            return 1
    else:
        # left subtree
        left = rcost(fam1,fam2,geneProximityD,proximityThreshold,subtree[1],rootProximate)
        chLeft = 1 + rcost(fam1,fam2,geneProximityD,proximityThreshold,subtree[1],not rootProximate)

        # right subtree
        right = rcost(fam1,fam2,geneProximityD,proximityThreshold,subtree[2],rootProximate)
        chRight = 1 + rcost(fam1,fam2,geneProximityD,proximityThreshold,subtree[2],not rootProximate)

        return min(left,chLeft) + min(right,chRight)

def costDiff(fam1,fam2,geneProximityD,proximityThreshold,subtree):
    '''Given two families calculate the difference in rcost depending on
whether we assume the root is not proximate or proximate. fam1 and
fam2 are family tuples specifying the genes present in a family.
    '''
    t=rcost(fam1,fam2,geneProximityD,proximityThreshold,subtree,True)
    f=rcost(fam1,fam2,geneProximityD,proximityThreshold,subtree,False)
    return(f-t)
    
def rscore(is0,is1,geneProximityD,proximityThreshold,subtree,familyL):
    '''Returns the rearrangement score between islands is0 and is1. Considers
all four ways these islands could join (since there are two islands each
with two ends). This is the rcost given we start not proximate minus
the rcost given we start proximate.
    '''
    if is0.mrca != is1.mrca:
        return -float('inf')
    else:
        # we try combining in each of the 4 orientations (but if island
        # only has 1 family, then we can skip the reverse orientation
        # for that one.

        caseL = [-float('inf')]*4

        # case 0: last fam in is0 vs. first fam in is1
        caseL[0] = costDiff(familyL[is0.familyL[-1]],familyL[is1.familyL[0]],geneProximityD,proximityThreshold,subtree)

        # case 1: last fam in is0 vs. last fam in is1
        if len(is1.familyL) == 1:
            caseL[1] = caseL[0] # since first and last of is1 are same
        else:
            caseL[1] = costDiff(familyL[is0.familyL[-1]],familyL[is1.familyL[-1]],geneProximityD,proximityThreshold,subtree)

        # case 2: first fam in is0 vs. first fam in is1
        if len(is0.familyL) == 1:
            caseL[2] = caseL[0] # since first and last of is0 are same
        else:
            caseL[2] = costDiff(familyL[is0.familyL[0]],familyL[is1.familyL[0]],geneProximityD,proximityThreshold,subtree)

        # case 3: first fam in is0 vs. last fam in is1
        if len(is0.familyL) == 1 and len(is1.familyL) > 1:
            caseL[3] = caseL[1]
        elif len(is0.familyL) > 1 and len(is1.familyL) == 1:
            caseL[3] = caseL[2]
        elif len(is0.familyL) == 1 and len(is1.familyL) == 1:
            caseL[3] = caseL[0]
        else: # both longer than 1
            caseL[3] = costDiff(familyL[is0.familyL[0]],familyL[is1.familyL[-1]],geneProximityD,proximityThreshold,subtree)

        return tuple(caseL)

def storeScore(is0,is1,geneProximityD,proximityThreshold,subtree,scoreD,familyL):
    '''Calculate and store score in scoreD. We follow convention that key
in scoreD should always have the lower island id first.'''
    if is0.id < is1.id:
        key = is0.id,is1.id
        tempScoreT=rscore(is0,is1,geneProximityD,proximityThreshold,subtree,familyL)
        # rscore returns different things depending on order
        # so we must be consistent with what we do in key.
    else:
        key = is1.id,is0.id
        tempScoreT=rscore(is1,is0,geneProximityD,proximityThreshold,subtree,familyL)
        
    scoreD[key]=(max(tempScoreT),tempScoreT)

    
def createScoreD(islandL,geneProximityD,proximityThreshold,subtree,familyL):
    '''Create dictionary of scores between all islands at a single node
(and thus with the same mrca).'''
    scoreD={}
    for i in range(len(islandL)-1):
        for j in range(i+1,len(islandL)):
            storeScore(islandL[i],islandL[j],geneProximityD,proximityThreshold,subtree,scoreD,familyL)
    return scoreD


## Iterative merging

def maxScore(scoreD):
    '''Find the entry with the highest score in scoreD, and return the
corresponding key.'''
    bestSc=-float('inf')
    bestIslandPairT=()
    bestScoreT=()
    for islandPairT,(sc,scoreT) in scoreD.items():
        if sc > bestSc:
            bestSc=sc
            bestIslandPairT=islandPairT
            bestScoreT=scoreT
    return bestSc,bestIslandPairT,bestScoreT

def delScores(scoreD,is0ID,is1ID):
    '''Given two island ids from newly merged nodes, delete any entries in
score D that come from them.'''
    # find ones to delete
    toDelL=[]
    for key in scoreD.keys():
        if is0ID in key or is1ID in key:
            toDelL.append(key)

    # now delete them
    for key in toDelL:
        del scoreD[key]

def addScores(scoreD,is0,islandsAtThisNodeL,geneProximityD,proximityThreshold,subtree,familyL):
    '''Get scores for island is0 against all other islands and add to
scoreD.'''
    for isl in islandsAtThisNodeL:
        if isl.id != is0.id:
            storeScore(is0,isl,geneProximityD,proximityThreshold,subtree,scoreD,familyL)

def searchIslandsByID(listOfIslands,id):
    '''Search for an island with id equal to id. Return the index of the
first we find, and None if their isn't one.
    '''
    for i,isl in enumerate(listOfIslands):
        if isl.id == id: return i,isl
    return None,None
    
def mergeIslandsAtNode(argT):
    '''Given a list of islands at one node (islandL) iteratively merge until
there are no more pairwise scores above threshold. We consider pairs
of (proximity threshold, rscore threshold) which are specified in the
list proxThreshL inside argT. So we might first do all merging for
proximity threshold 1 (adjacent genes), merging until there are no
islands whose pairwise rscore is >= to rscore threshold. Then we
proceed to the next (proximity threshold, rscore threshold) pair,
which might have for example proximity threshold 2 (we consider genes
nearby that are separated by 1 gene).
    '''

    islandL,geneProximityD,proxThreshL,subtree,familyL = argT
    
    if len(islandL) < 2:
        # nothing to merge
        return islandL

    # we consider every pair of proximity threshold and rscore
    # threshold in proxThreshL
    for proximityThreshold,rscThreshold in proxThreshL:

        scoreD = createScoreD(islandL,geneProximityD,proximityThreshold,subtree,familyL)
        
        while True:
            sc,islandPairT,scoreT=maxScore(scoreD)
            if sc < rscThreshold:
                break

            ind0,is0 = searchIslandsByID(islandL,islandPairT[0])
            ind1,is1 = searchIslandsByID(islandL,islandPairT[1])

            is0.merge(is1,scoreT.index(sc))

            # delete is1
            del islandL[ind1]

            # remove all scores in scoreD that involve is0 or is1
            delScores(scoreD,is0.id,is1.id)

            # calculate new scores for is0 against all other islands
            addScores(scoreD,is0,islandL,geneProximityD,proximityThreshold,subtree,familyL)

    return islandL

## Output

def printSummary(islandByNodeLMerged,focalNodesL,strainNum2StrD,islandByNodeL,outputSummaryF):
    '''Print a summary of the merging saying how many islands there were
at each node in the focal clade, before and after merging.'''

    focalPrintD = {}
    for islandGroup in islandByNodeLMerged:
        if islandGroup != []:
            mrca = islandGroup[0].mrca
            if mrca in focalNodesL:
                focalPrintD[mrca] = [strainNum2StrD[mrca],str(len(islandByNodeL[mrca])),str(len(islandGroup))]

    print("Number of islands per node in focal clade: ",file=outputSummaryF)
    rowL=[]
    rowL.append(['Node','Before merge','After merge'])
    rowL.append(['----','------------','-----------'])
    for node in focalNodesL:
        if node in focalPrintD:
            rowL.append(focalPrintD[node])
        else:
            rowL.append([strainNum2StrD[node],'0','0'])

    analysis.printTable(rowL,indent=2,fileF=outputSummaryF)

    return


def writeIslands(islandByNodeL,strainNum2StrD,islandOutFN):
    '''Write the islands to a file'''
    f=open(islandOutFN,"w")
    for branch in range(len(islandByNodeL)):
        for island in islandByNodeL[branch]:
            print(island.fileStr(strainNum2StrD),file=f)
    f.close()


def readIslands(islandFN,tree,strainStr2NumD):
    '''Given a file name for a islands output file, load back
recreating islandByNodeL.'''

    islandByNodeL=[[] for i in range(trees.nodeCount(tree))]
    
    f=open(islandFN,'r')
    while True:
        s=f.readline()
        if s == '':
            break
        gr=str2Island(s.rstrip(),strainStr2NumD)
        islandByNodeL[gr.mrca].append(gr)
    f.close()

    return islandByNodeL



import numpy,glob,struct

class Score:

    def __init__(self):
        '''Create an object for storing scores.'''
        
        self.endNodesToEdgeD = {}
        self.scoreD = {}

    def initializeDataAttributes(self,blastFilePath,geneNames):
        '''This method takes a new, empty object and fills the data attributes
by reading through blast files to identify pairs of genes with
edges. Also creates the arrays for storing scores, initialized to 0.'''
        self.fillEndNodesToEdgeD(blastFilePath,geneNames)
        self.numEdges=len(self.endNodesToEdgeD)
        self.initializeScoreD()

    def fillEndNodesToEdgeD(self,blastFilePath,geneNames):
        '''Run through blast files, finding all pairs of genes with signicant similarity. Use these to fill endNodesToEdgeD.'''

        blastFnL=glob.glob(blastFilePath)
        
        # Run through blast files getting list of all genes to compare. The
        # blast files have redundancies (ie , g1,g2 and also g2,g1). Keep
        # comparisons to do in set, and don't add if we already have it in
        # either order.
        edgeNum=0
        for fn in blastFnL:
            f = open(fn,'r')
            while True:
                s = f.readline()
                if s=='':
                    break
                L = s.split('\t')
                if len(L) != 12: # we only want lines with 12 columns
                    continue

                g1 = geneNames.nameToNum(L[0])
                g2 = geneNames.nameToNum(L[1])

                # make sure g1 is the lower gene number (we do this so
                # we only store in one orientation in dict).
                if g1 > g2: g2,g1 = g1,g2

                if (g1,g2) not in self.endNodesToEdgeD:
                    # we haven't got it yet, add it
                    self.endNodesToEdgeD[(g1,g2)] = edgeNum
                    edgeNum += 1
            f.close()

    def initializeScoreD(self):
        '''Create the arrays for storing scores.'''
        
        self.scoreD['rawSc'] = numpy.zeros(self.numEdges,dtype=numpy.float64)
        self.scoreD['normSc'] = numpy.zeros(self.numEdges,dtype=numpy.float64)
        self.scoreD['synSc'] = numpy.zeros(self.numEdges,dtype=numpy.float64)
        self.scoreD['coreSynSc'] = numpy.zeros(self.numEdges,dtype=numpy.float64)

    def isEdgePresentByEndNodes(self,g1,g2):
        '''See if g1 and g2 have an edge in our data structure. Return
boolean.'''
        if g1 > g2: g2,g1 = g1,g2
        if (g1,g2) in self.endNodesToEdgeD: return True
        else: return False
        
    def endNodesToEdge(self,g1,g2):
        '''Given two genes, return the number of the edge between them. If
there isn't any, return None.'''
        edge = self.endNodesToEdgeD[(g1,g2)]
        return edge

    def addScoreByEdge(self,edge,sc,scoreType):
        '''Given a score on an edge, store it in the score array corresponding
to scoreType.
        '''
        self.scoreD[scoreType][edge] = sc
        
    def addScoreByEndNodes(self,g1,g2,sc,scoreType):
        '''Given a score between two genes store it in the score array
corresponding to scoreType.'''

        if g1 > g2: g2,g1 = g1,g2 # make sure g1 is lower gene num

        if self.isEdgePresentByEndNodes(g1,g2):
            # we can only add scores for those edges that already exist
            edge = self.endNodesToEdge(g1,g2)
            self.addScoreByEdge(edge,sc,scoreType)
        # if the edge isn't already present, do nothing.
            
    def getScoreByEdge(self,edge,scoreType):
        '''Given an edge get the score corresponding to scoreType.'''
        return self.scoreD[scoreType][edge]

    def getScoreByEndNodes(self,g1,g2,scoreType):
        '''Given two genes get the score corresponding to scoreType.'''

        if g1 > g2: g2,g1 = g1,g2 # make sure g1 is lower gene num
        edge = self.endNodesToEdge(g1,g2)
        return self.getScoreByEdge(edge,scoreType)

    def iterateEdges(self):
        '''Returns an iterator which goes over edges by edge number.'''
        return range(self.numEdges)

    def iterateEdgesByEndNodes(self):
        '''Returns an iterator which goes over edges by end nodes.'''
        return self.endNodesToEdgeD.keys()

    def createNodeConnectL(self,geneNames):
        '''Create an attribute nodeConnectL. Index in this list corresponds to
gene. Value at that index is a list of the genes which a given gene
connects to. This attribute is not saved in our file formats. It must
be recalculated before it will be used (e.g. in family formation).'''

        self.nodeConnectL = [[] for gn in geneNames.nums]

        # loop over endNodesToEdgeD populating nodeConnectL
        for gn1,gn2 in self.endNodesToEdgeD:
            self.nodeConnectL[gn1].append(gn2)
            self.nodeConnectL[gn2].append(gn1)

    def getConnectionsGene(self,gene):
        '''Return a list containing all the genes connected to gene.'''
        return self.nodeConnectL[gene]

    def createNodeEdgeL(self,geneNames):
        '''Create an attribute nodeEdgeL. Index in this list corresponds to
gene. Value at that index is a list of the edges which a give gene
connects to. This attribute is not saved in our file formats. It must
be recalculated before it will be used.'''

        self.nodeEdgeL = [[] for gn in geneNames.nums]

        # loop over endNodesToEdgeD populating nodeEdgeL
        for gn1,gn2 in self.endNodesToEdgeD:
            edge = self.endNodesToEdge(gn1,gn2)
            self.nodeEdgeL[gn1].append(edge)
            self.nodeEdgeL[gn2].append(edge)

    def getConnectionsEdge(self,gene):
        '''Return a list containing all the edges connected to gene.'''
        return self.nodeEdgeL[gene]

    def createEdgeToEndNodeL(self):
        '''Create an attribute edgeToEndNodeL where the index is edge number,
and the value is a tuple giving the two genes on either end of the
edge. This attribute is not saved in our file formats. It must be
recalculated before it will be used.
        '''
        self.edgeToEndNodeL = [None for x in range(len(self.endNodesToEdgeD))]
        
        for endPairT in self.endNodesToEdgeD:
            edge = self.endNodesToEdgeD[endPairT]
            self.edgeToEndNodeL[edge] = endPairT

    def getEndNodesByEdge(self,edge):
        '''Given and edge, return the numbers for the two genes on either end.'''
        return self.edgeToEndNodeL[edge]
        
    def __eq__(self,other):
        '''Compare two Score objects to see if they have the same values.'''
        
        c1 = self.endNodesToEdgeD == other.endNodesToEdgeD
        c2 = all(self.scoreD['rawSc'] == other.scoreD['rawSc'])
        c3 = all(self.scoreD['normSc'] == other.scoreD['normSc'])
        c4 = all(self.scoreD['synSc'] == other.scoreD['synSc'])
        c5 = all(self.scoreD['coreSynSc'] == other.scoreD['coreSynSc'])

        if c1 and c2 and c3 and c4 and c5:
            return True
        else:
            return False
        
    def writeScoresText(self,geneNames,scoresFN):
        '''Save all edges (pairs of genes) to a tab delimited text file with a
    header line.
        '''

        scoreTypeL = ['rawSc','normSc','synSc','coreSynSc']

        # open file
        f=open(scoresFN,'w')

        # write header
        f.write("\t".join(['gene1','gene2','edge']+scoreTypeL)+'\n')

        for gene1Num,gene2Num in self.endNodesToEdgeD:
            edge = self.endNodesToEdgeD[(gene1Num,gene2Num)]
            outStrL=[]
            outStrL.append(geneNames.numToName(gene1Num))
            outStrL.append(geneNames.numToName(gene2Num))
            outStrL.append(str(edge))
            for scoreType in scoreTypeL:
                outStrL.append(format(self.getScoreByEndNodes(gene1Num,gene2Num,scoreType),".6f"))

            f.write("\t".join(outStrL)+'\n')

        f.close()

    def writeScoresBinary(self,scoresFN):
        '''Write scores to scoresFN as a binary file.'''

        scoreTypeL = ['rawSc','normSc','synSc','coreSynSc']
        f=open(scoresFN,'wb')

        # first 8 bytes are int, telling us how many edges
        f.write(self.numEdges.to_bytes(8,'little'))

        # write a block of bytes for each edge.
        # g1 g2 edge
        for g1,g2 in self.endNodesToEdgeD:
            edge = self.endNodesToEdgeD[(g1,g2)]
            f.write(g1.to_bytes(8,'little'))
            f.write(g2.to_bytes(8,'little'))
            f.write(edge.to_bytes(8,'little'))
            for scoreType in scoreTypeL:
                val = float(self.getScoreByEndNodes(g1,g2,scoreType))
                f.write(struct.pack('<d',val))
        
        f.close()
        
    def readScoresText(scoresFN,geneNames):
        '''Read scores from a text file of scores and use to create Score
        object
        '''

        scoresO = Score()
        
        # first pass over file, fill endNodesToEdgeD
        f = open(scoresFN,'r')
        f.readline() # skip header
        while True:
            s = f.readline()
            if s == '':
                break
            lineL=s.split('\t')
            g1 = geneNames.nameToNum(lineL[0])
            g2 = geneNames.nameToNum(lineL[1])
            edge = int(lineL[2])
            scoresO.endNodesToEdgeD[(g1,g2)] = edge
        f.close()
        
        # Set up the arrays
        scoresO.numEdges=len(scoresO.endNodesToEdgeD)
        scoresO.initializeScoreD()


        # 2nd pass, fill arrays
        f = open(scoresFN,'r')
        headerL=f.readline().split()
        scoreTypeL=headerL[3:] # first 3 cols are gene names + edge

        while True:
            s = f.readline()
            if s == '':
                break
            lineL=s.split('\t')
            edge = int(lineL[2])
            scoreL=lineL[3:]
            for i,sc in enumerate(scoreL):
                scoresO.scoreD[scoreTypeL[i]][edge] = float(sc)
        f.close()
        return scoresO

    def readScoresBinary(scoresFN):
        '''Read scores from a binary files.'''

        scoresO=Score()
        f=open(scoresFN,'rb')

        # read first 8 bytes to get numEdges
        b = f.read(8)
        scoresO.numEdges = int.from_bytes(b,'little')

        # initialize score arrays
        scoresO.initializeScoreD()
        
        # loop over blocks of bytes where each block is an edge
        for i in range(scoresO.numEdges):
            g1 = int.from_bytes(f.read(8),'little')
            g2 = int.from_bytes(f.read(8),'little')
            edge = int.from_bytes(f.read(8),'little')
            
            scoresO.endNodesToEdgeD[(g1,g2)] = edge

            # fill array
            scoresO.scoreD['rawSc'][edge] = struct.unpack('<d',f.read(8))[0]
            scoresO.scoreD['normSc'][edge] = struct.unpack('<d',f.read(8))[0]
            scoresO.scoreD['synSc'][edge] = struct.unpack('<d',f.read(8))[0]
            scoresO.scoreD['coreSynSc'][edge] = struct.unpack('<d',f.read(8))[0]            
            
        f.close()
        return scoresO




class Family:

    def __init__(self, idnum, mrca, genesL,numNodesInTree,geneNames):
        '''Given an family ID number, a most recent common ancestor, and a
list of genes create a family object. Handles genes specified either
numerically or by name. numNodesInTree is the number of nodes found in
the tree (the full tree) we're working with for the
project. possibleErrorCt is a count of near miss genes, that either
were added but almost weren't, or were not added but almost were.
        '''

        self.id = idnum
        self.mrca = mrca

        # create the family gene tuple, This has indexes corresponding
        # to nodes on the tree. At each position we have the genes at
        # that node.

        famGeneL = [[] for j in range(numNodesInTree)]
        geneNumL=[]
        for gene in genesL:
            # if we're being passed in genes with string names,
            # convert to number
            if type(gene)==int:
                geneNum=gene
            else:
                geneNum=geneNames.nameToNum(gene)

            strainNum=geneNames.numToStrainNum(geneNum)
            famGeneL[strainNum].append(geneNum)

        # tuple-ize
        newFamGeneL=[]
        for L in famGeneL:
            newFamGeneL.append(tuple(L))
        
        self.famGeneT = tuple(newFamGeneL)

    def getGeneNums(self):
        '''Extract and return all the gene numbers from this family.'''
        genesL=[]
        for geneT in self.famGeneT:
            genesL.extend(geneT)
        return genesL

    def getGeneNames(self,geneNames):
        '''Extract and return all the gene names from this family.'''
        genesL=[]
        for num in self.getGeneNums():
            genesL.append(geneNames.numToName(num))
        return genesL

    def isInStrain(self,strainNum):
        '''Is this family present in strainNum, return boolean.'''
        if len(self.famGeneT[strainNum]) > 0:
            return True
        else: return False
        
    def __repr__(self):
        '''String representation of a family containing family number.'''
        return "<Family: "+str(self.id) + ">"

    def fileStr(self,geneNames,strainNum2StrD):
        '''String representation suitable for writing the family to
file. Genes and mrca are expressed in word form.'''

        outL =[str(self.id)]
        outL.append(strainNum2StrD[self.mrca])
        outL.extend(self.getGeneNames(geneNames))
        return "\t".join(outL)

    def getOutsideConnections(self,scoresO):
        '''Given a score object, return a tuple of all outside genes with
connections to this family.'''
        allGenesInFamL = self.getGeneNums()
        otherGenesS=set()
        for geneNum in allGenesInFamL:
            for otherGene in scoresO.getConnectionsGene(geneNum):
                if not otherGene in allGenesInFamL:
                    otherGenesS.add(otherGene)
        return tuple(otherGenesS)
    
    def getPossibleErrorCt(self, scoresO,minNormThresh,minCoreSynThresh,minSynThresh,famErrorScoreIncrementD):
        '''Given a family, returns the number of near misses internally and externally'''
        internalPossibleErrors,externalPossibleErrors = 0,0
        allGenesInFamL = self.getGeneNums()
        internalEdgeL = makeMSN(allGenesInFamL,scoresO)
        externalGenesTuple=self.getOutsideConnections(scoresO)
        externalEdgeL = makeExternalEdgeL(externalGenesTuple, allGenesInFamL, scoresO)
        for g1,g2 in internalEdgeL:
            internalPossibleErrors+=isPossibleErrorInternal(g1,g2,scoresO,minNormThresh,minCoreSynThresh,minSynThresh,famErrorScoreIncrementD)
        for g1,g2 in externalEdgeL:
            externalPossibleErrors+=isPossibleErrorExternal(g1,g2,scoresO,minNormThresh,minCoreSynThresh,minSynThresh,famErrorScoreIncrementD)

        self.possibleErrorCt = internalPossibleErrors + externalPossibleErrors


# Functions to help in calculating possible error count

def makeExternalEdgeL(externalGenesTuple, genesInFam, scoresO):
    '''given a family, returns a list of tuples '''
    returnEdgeL = [0]*len(externalGenesTuple)
    for extGeneIndex in range(0,len(externalGenesTuple)):
        bestScore = float('-inf')
        extGene = externalGenesTuple[extGeneIndex]
        connectedGenes = scoresO.getConnectionsGene(extGene)
        for conGene in connectedGenes:
            if conGene in genesInFam:
                rawSc=scoresO.getScoreByEndNodes(extGene,conGene, 'rawSc')
                if rawSc>bestScore:
                    bestScore=rawSc
                    returnEdgeL[extGeneIndex]=(extGene,conGene)
    return returnEdgeL

def makeInternalEdgeL(genesInFam,scoresO):
    '''Given a list of all genes in the family, it returns a
list of lists sorted in descending order by raw score'''
    # compare every gene to every other gene that comes
    # in the list after it
    # format: [[rawScore, gene1, gene2], ....]] for every edge in the fam
    totalGenesNum = len(genesInFam)
    internalEdgeL = []
    for gene1Index in range(0,totalGenesNum-1):
        for gene2Index in range(gene1Index+1,totalGenesNum):
            g1=genesInFam[gene1Index]
            g2=genesInFam[gene2Index]
            if scoresO.isEdgePresentByEndNodes(g1,g2):
                rawSc = scoresO.getScoreByEndNodes(g1,g2, 'rawSc')
                internalEdgeL.append([rawSc,g1,g2])
    internalEdgeL.sort(reverse=True)
    return internalEdgeL


def makeUnconnectedGsetL(genesInFam):
    '''Given a list of all genes in the family, make a set of disjoint genes'''
    unconnectedGsetL = []
    for gene in genesInFam:
        geneSet = set()
        geneSet.add(gene)
        unconnectedGsetL.append(geneSet)
    return unconnectedGsetL


def updateConnections(gSetL, g1, g2):
    '''Given 2 genes and a gene set list, returns True if the genes
are disconnected, false otherwise. If genes are disconnected, join the
genes and update gSetL accordingly'''
    for setIndex1 in range(0,len(gSetL)):
        if g1 in gSetL[setIndex1]:
               if g2 in gSetL[setIndex1]: return False,gSetL
               else:
                   for setIndex2 in range(0, len(gSetL)):
                       if g2 in gSetL[setIndex2]:
                           gSetL[setIndex1]=gSetL[setIndex1].union(gSetL[setIndex2])
                           gSetL.remove(gSetL[setIndex2])
                           return True, gSetL
    return False, gSetL
        

def makeMSN(genesInFam,scoresO):
    '''Given a tuple of genes in the family, it makes the max spanning network based on
the largest raw scores'''
    returnEdgeList = []
    internalEdgeL = makeInternalEdgeL(genesInFam,scoresO)
    gSetL = makeUnconnectedGsetL(genesInFam)
    while len(gSetL) > 1:
        currEdge = internalEdgeL[0]
        internalEdgeL.remove(currEdge)
        g1,g2 = currEdge[1],currEdge[2]
        isDisconnected, gSetL = updateConnections(gSetL, g1, g2)
        if isDisconnected:
            returnEdgeList.append((g1,g2))
    return returnEdgeList

def isPossibleErrorInternal(g1,g2,scoresO,minNormThresh,minCoreSynThresh,minSynThresh,famErrorScoreIncrementD):
    '''Given g1 and g2 both inside a family, check the various scores to
    determine if it was a near miss, ie almost was not in the
    family. Returns boolean.
    '''
    # get scores
    normSc = scoresO.getScoreByEndNodes(g1,g2,'normSc')
    synSc = scoresO.getScoreByEndNodes(g1,g2,'synSc')
    coreSynSc = scoresO.getScoreByEndNodes(g1,g2,'coreSynSc')
    
    if not (normSc - famErrorScoreIncrementD['normSc']) >= minNormThresh:
        # Was over, but now is below threshold. Is a possible error
        return True
    elif not (coreSynSc - famErrorScoreIncrementD['coreSynSc']) >= minCoreSynThresh:
        return True
    elif not (synSc - famErrorScoreIncrementD['synSc']) >= minSynThresh:
        return True
    else: return False

def isPossibleErrorExternal(g1,g2,scoresO,minNormThresh,minCoreSynThresh,minSynThresh,famErrorScoreIncrementD):
    '''Given g1 inside and g2 outside a family, check the various scores
    to determine if it was a near miss, ie almost was put in the
    family. We consider normSc, synSc, and coreSynSc. We define near
    miss to be cases when 2/3 of these were above threshold, and the
    third was below, but within an increment of the threshold
    (increments provided in famErrorScoreIncrementD). Returns boolean.
    '''
    # get scores
    normSc = scoresO.getScoreByEndNodes(g1,g2,'normSc')
    synSc = scoresO.getScoreByEndNodes(g1,g2,'synSc')
    coreSynSc = scoresO.getScoreByEndNodes(g1,g2,'coreSynSc')

    # w/out increment
    normB = normSc >= minNormThresh
    coreSynB = coreSynSc >= minCoreSynThresh
    synB = synSc >= minSynThresh

    # with increment
    normIncB = (normSc + famErrorScoreIncrementD['normSc']) >= minNormThresh
    coreSynIncB = (coreSynSc + famErrorScoreIncrementD['coreSynSc']) >= minCoreSynThresh
    synIncB = (synSc + famErrorScoreIncrementD['synSc']) >= minSynThresh
    
    if coreSynB and synB and not normB and normIncB:
        # core and syn were over without increment, but norm is only
        # over once we add the increment
        return True
    elif normB and synB and not coreSynB and coreSynIncB:
        return True
    elif coreSynB and normB and not synB and synIncB:
        return True
    else: return False



# Functions for a modiefied version of the PhiGs algorithm
# http://www.ncbi.nlm.nih.gov/pmc/articles/PMC1523372/
# (We've added the use of synteny)
import sys
from . import trees
from .Family import *


## Main function

def families(tree,subtreeL,geneNames,scoresO,minNormThresh,minCoreSynThresh,minSynThresh,synAdjustThresh,synAdjustExtent,familyFN,strainNum2StrD, outputSummaryF):
    '''Given a graph of genes and their similarity scores find families
using a PhiGs-like algorithm, with synteny also considered.
    '''

    nodeOrderL=createNodeProcessOrderList(tree)

    # initialize scoresO.nodeConnectL for use below
    scoresO.createNodeConnectL(geneNames)
    
    geneUsedL = [False for x in geneNames.nums]
    multiGeneFamilyL=[]
    
    for node,lnode,rnode in nodeOrderL:
        if lnode != None:
            # not a tip
            subtree=subtreeL[node]
            leftS,rightS,outgroupS = createLRSets(subtree,geneNames)
            seedL = createSeedL(leftS,rightS,scoresO,minNormThresh,minCoreSynThresh,minSynThresh)
            for seed in seedL:
                seedRawSc,g1,g2 = seed

                if seedRawSc == -float('inf'):
                    # we've gotten to the point in the seed list with
                    # genes having no match on the other branch
                    break
                else:
                    famS=getFamily(seedRawSc,g1,g2,scoresO,leftS,rightS,minNormThresh,minCoreSynThresh,minSynThresh,synAdjustThresh,synAdjustExtent,geneUsedL)
                    
                    if famS == None:
                        # one of the genes the the family was already
                        # used, so getFamily returned None
                        continue
                    else:
                        # none of the genes in famS used yet
                        for gene in famS:
                            geneUsedL[gene] = True
                        multiGeneFamilyL.append((node,famS))

    # Create Family objects 
    genesInMultiGeneFamsS = set()
    familyL=[]
    famNum = 0
    for mrca,famS in multiGeneFamilyL:
        genesInMultiGeneFamsS.update(famS) # add all genes in fam
        familyL.append(Family(famNum,mrca,famS,trees.nodeCount(tree),geneNames ))
        famNum+=1

    multiGeneFamNum=famNum
    print("Number of multigene families",multiGeneFamNum,file=outputSummaryF)

    # Add on remaining genes as single gene families.
    for gene in geneNames.nums: 
        if not gene in genesInMultiGeneFamsS:
            mrca = geneNames.numToStrainNum(gene)
            familyL.append(Family(famNum,mrca,[gene],trees.nodeCount(tree),geneNames))
            famNum+=1

    print("Number of single gene families",famNum-multiGeneFamNum,file=outputSummaryF)
    print("Number of total families",famNum,file=outputSummaryF)

    writeFamilies(familyL,geneNames,strainNum2StrD,familyFN)

    return tuple(familyL)

## Support functions

def createNodeProcessOrderList(tree):
    '''Given a tree, output a list specifying nodes in pre-order (ancestors
before their descendants). For each node we give a tuple (node #, left
node #, right node #).
    '''
    if tree[1] == ():
        return [(tree[0], None, None)]
    else:
        l = createNodeProcessOrderList(tree[1])
        r = createNodeProcessOrderList(tree[2])
        return [(tree[0], tree[1][0], tree[2][0])] + l + r


def createLRSets(tree,geneNames):
    '''For every gene in our data, put it into one of three sets. Left,
right, or outgroup. Genes in the left set are found in a species on
the left branch of tree.'''

    leftSpeciesS=set(trees.leafList(tree[1]))
    rightSpeciesS=set(trees.leafList(tree[2]))
    
    leftS=set()
    rightS=set()
    outgroupS=set()
    for geneNum in geneNames.nums: # all genes
        strain=geneNames.numToStrainNum(geneNum)
        if strain in leftSpeciesS:
            leftS.add(geneNum)
        elif strain in rightSpeciesS:
            rightS.add(geneNum)
        else:
            outgroupS.add(geneNum)

    return(leftS,rightS,outgroupS)

def closestMatch(gene,S,scoresO,minNormThresh,minCoreSynThresh,minSynThresh):
    '''Find the closest match to gene among the genes in the set S in the
graph scoresO. Eliminate any matches that have a norm score below
minNormThresh, a coreSynSc below minCoreSynThresh, or a synteny score
below minSynThresh.
    '''
    bestGene=None
    bestEdgeScore = -float('inf')
    for otherGene in scoresO.getConnectionsGene(gene):
        if otherGene in S:
            if scoresO.getScoreByEndNodes(gene,otherGene,'rawSc') > bestEdgeScore:
                if scoresO.getScoreByEndNodes(gene,otherGene,'normSc') >= minNormThresh:
                    if scoresO.getScoreByEndNodes(gene,otherGene,'coreSynSc') >= minCoreSynThresh:
                        if scoresO.getScoreByEndNodes(gene,otherGene,'synSc') >= minSynThresh:
                            bestEdgeScore = scoresO.getScoreByEndNodes(gene,otherGene,'rawSc')
                            bestGene = otherGene
    return bestEdgeScore, gene, bestGene
    
def createSeedL(leftS,rightS,scoresO,minNormThresh,minCoreSynThresh,minSynThresh):
    '''Create a list which has the closest match for each gene on the
opposite side of the tree. e.g. if a gene is in tree[1] then we're
looking for the gene in tree[2] with the closest match. We eliminate
any matches that are below threshold for normalized or syntenty
scores. For each gene we get this closest match, put in a list, sort,
and return.
    '''
    seedL=[]
    for gene in leftS:
        seedL.append(closestMatch(gene,rightS,scoresO,minNormThresh,minCoreSynThresh,minSynThresh))
    for gene in rightS:
        seedL.append(closestMatch(gene,leftS,scoresO,minNormThresh,minCoreSynThresh,minSynThresh))
    seedL.sort(reverse=True)
    return seedL

def getFamily(seedRawSc,g1,g2,scoresO,leftS,rightS,minNormThresh,minCoreSynThresh,minSynThresh,synAdjustThresh,synAdjustExtent,geneUsedL):
    '''Based on a seed (seedScore, g1, g2) search for a family. Using the
PhiGs approach, we collect all genes which are closer to members of
the family than the two seeds are from each other. We have a
normScore, synScore, and coreSynScore thresholds below which we will
not add genes. We also have a synteny adjustment of the raw score
where we make the raw score between a pair a bit better if their
synteny is above synAdjustThresh. In general, if a gene has syntenic
connectsions to genes already in the family, this makes us more
confident that this gene belongs in the family. Returns a set
containing genes in the family.
    '''

    if geneUsedL[g1] or geneUsedL[g2]:
        # one of these has been used already, stop now.
        return None

    alreadySearchedS = set()
    notYetSearchedS = set([g1,g2])

    while len(notYetSearchedS) > 0:
        newGenesS=set()
        for famGene in notYetSearchedS:
            for newGene in scoresO.getConnectionsGene(famGene):
                if newGene in leftS or newGene in rightS:
                    # it is from a species descended from the node
                    # we're working on
                    if newGene not in alreadySearchedS and newGene not in newGenesS and newGene not in notYetSearchedS:
                        addIt = addGene(famGene,newGene,scoresO,minNormThresh,minCoreSynThresh,minSynThresh,seedRawSc,synAdjustThresh,synAdjustExtent)
                        if addIt:
                            if geneUsedL[newGene]:
                                # this one's been used already. That
                                # means the whole family should be
                                # thrown out. Just stop now.
                                return None
                            else:
                                newGenesS.add(newGene)
                    
            alreadySearchedS.add(famGene)
        notYetSearchedS = newGenesS
    return alreadySearchedS


def addGene(famGene,newGene,scoresO,minNormThresh,minCoreSynThresh,minSynThresh,seedRawSc,synAdjustThresh,synAdjustExtent):
    '''Given famGene that is inside a family, and newGene we are
considering adding, check the various scores to determine if we should
add it. Return boolean.
    '''
    addIt = False
    
    rawSc = scoresO.getScoreByEndNodes(famGene,newGene,'rawSc')
    normSc = scoresO.getScoreByEndNodes(famGene,newGene,'normSc')
    synSc = scoresO.getScoreByEndNodes(famGene,newGene,'synSc')
    coreSynSc = scoresO.getScoreByEndNodes(famGene,newGene,'coreSynSc')

    if normSc < minNormThresh:
        # doesn't meet score threshold for family
        # formation. (Modification of PhiGs)
        pass
    elif coreSynSc < minCoreSynThresh:
        # doesn't meet min core synteny requirement for
        # family formation. (Modification of PhiGs)
        pass
    elif synSc < minSynThresh:
        # doesn't meet min synteny requirement for
        # family formation. (Modification of PhiGs)
        pass
    elif rawSc >= seedRawSc:
        # If its within the seed distance, add it
        # (basic PhiGs approach). we have the =
        # there in case seedRawSc is 1.
        addIt = True
    elif synSc > synAdjustThresh:
        # its above the syn score adjustment
        # threshold, so increase rawSc a bit. This
        # addresses a problem with closely related
        # families where the seed score is very
        # similar. Sometimes by chance things
        # that should have been added weren't
        # because they weren't more similar than
        # an already very similar seed.
        # (Modification of PhiGs)
        adjSc = rawSc * synAdjustExtent
        if adjSc > 1: adjSc = 1 # truncate back to 1
        if adjSc >= seedRawSc:
            addIt = True
    return addIt


def calcErrorScores(familyL,scoresO,minNormThresh,minCoreSynThresh,minSynThresh,famErrorScoreIncrementD):
    '''Given a list of family objects, calculate error scores for
each. These values get saved as attributes of the objects.
    '''
    for famO in familyL:
        famO.getPossibleErrorCt(scoresO,minNormThresh,minCoreSynThresh,minSynThresh,famErrorScoreIncrementD)


## Input/output

def writeFamilies(familyL,geneNames,strainNum2StrD,fileName):
    '''Write all gene families to fileName, one family per line.'''
    f=open(fileName,'w')
    for fam in familyL:
        f.write(fam.fileStr(geneNames,strainNum2StrD)+'\n')
    f.close()


def readFamilies(familyFN,tree,geneNames,strainStr2NumD):
    '''Read the family file named familyFN, creating a tuple of family
objects where the index corresponds to family number.'''
    
    famL=[]
    maxFamNum=0
    f=open(familyFN,'r')
    while True:
        s=f.readline()
        if s=='':
            break
        L=s.split()
        famNum=int(L[0])
        mrca = strainStr2NumD[L[1]]
        genesL = L[2:]

        famL.append(Family(famNum,mrca,genesL,trees.nodeCount(tree),geneNames))
    f.close()
    return famL




class Island:

    def __init__(self, idnum, mrca, familyL):
        '''Create a island object.'''
        self.id = idnum
        self.mrca = mrca
        self.familyL=familyL # list of family numbers in the island, in order
            
    def __repr__(self):
        return "<id:"+str(self.id)+", mrca:"+str(self.mrca)+", familyL:"+str(self.familyL)+">"

    def __len__(self):
        return len(self.familyL)
    
    def fileStr(self,strainNum2StrD):
        '''Return a string which can be used for saving a island compactly in a
file.'''
        return str(self.id)+"\t"+strainNum2StrD[self.mrca]+"\t"+",".join(map(str,self.familyL))
        
    def merge(self,other,orientation):
        '''Merge island other into self. The argument orientation tells us
which orientation to combine the families of self and other in. The
meaning of the different values is determined by the cases in the
score function.
        '''
        if orientation == 0:
            newFamilyL= self.familyL + other.familyL
        elif orientation == 1:
            newFamilyL= self.familyL + other.familyL[::-1]
        elif orientation == 2:
            newFamilyL= other.familyL[::-1] + self.familyL
        elif orientation == 3:
            newFamilyL= other.familyL + self.familyL
        self.familyL=newFamilyL
            
def str2Island(islandStr,strainStr2NumD):
    '''Given a island string (e.g. produced by the fileStr method) parse to produce a island.'''
    L=islandStr.split('\t')
    id=int(L[0])
    mrca=strainStr2NumD[L[1]]
    familyL=[int(x) for x in L[2].split(',')]
    return Island(id,mrca,familyL)
